@@ -4,11 +4,13 @@ import sys
 from collections import defaultdict
 from datetime import date, datetime
 from functools import wraps
+from typing import Any, Callable, DefaultDict, List, Optional, Sequence, Union
 
 import click
+import requests
 
 from . import __version__
-from .cve_api import CveApi, CveApiError
+from .cve_api import CveApi
 
 CVE_RE = re.compile(r"^CVE-[12]\d{3}-\d{4,}$")
 CONTEXT_SETTINGS = {
@@ -17,28 +19,30 @@ CONTEXT_SETTINGS = {
 }
 
 
-def validate_cve(ctx, param, value):
+def validate_cve(ctx: click.Context, param: click.Parameter, value: Optional[str]) -> Optional[str]:
     if value is None:
-        return
+        return None
     if not CVE_RE.match(value):
         raise click.BadParameter("invalid CVE ID.")
     return value
 
 
-def validate_year(ctx, param, value):
+def validate_year(
+    ctx: click.Context, param: click.Parameter, value: Optional[str]
+) -> Optional[str]:
     if value is None:
-        return
+        return None
     # Hopefully this code won't be around in year 10,000.
     if not re.match(r"^[1-9]\d{3}$", value):
         raise click.BadParameter("invalid year.")
     return value
 
 
-def human_ts(ts):
+def human_ts(ts: str) -> str:
     return datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%fZ").strftime("%c")
 
 
-def print_cve(cve):
+def print_cve_id(cve: dict) -> None:
     click.secho(cve["cve_id"], bold=True)
     click.echo(f"├─ State:\t{cve['state']}")
     # CVEs reserved by other CNAs do not include information on who requested them and when.
@@ -50,7 +54,14 @@ def print_cve(cve):
         click.echo(f"└─ Owning CNA:\t{cve['owning_cna']}")
 
 
-def print_table(lines):
+def print_cve_record(cve: dict) -> None:
+    click.secho(cve["cveMetadata"]["cveId"], bold=True)
+    click.echo(f"├─ State:\t{cve['cveMetadata']['state']}")
+    click.echo(f"├─ Owning CNA:\t{cve['cveMetadata']['assignerShortName']}")
+    click.echo(f"└─ Reserved on:\t{human_ts(cve['cveMetadata']['dateReserved'])}")
+
+
+def print_table(lines: Sequence[Sequence[str]]) -> None:
     """Print tabulated data based on the widths of the longest values in each column."""
     col_widths = []
     for item_index in range(len(lines[0])):
@@ -65,11 +76,11 @@ def print_table(lines):
             click.echo(text)
 
 
-def print_json_data(data):
+def print_json_data(data: Union[dict, list]) -> None:
     click.echo(json.dumps(data, indent=4, sort_keys=True))
 
 
-def print_user(user):
+def print_user(user: dict) -> None:
     name = get_full_name(user)
     if name:
         click.echo(f"{name} — ", nl=False)
@@ -85,47 +96,59 @@ def print_user(user):
     click.echo(f"└─ Modified:\t{human_ts(user['time']['modified'])}")
 
 
-def get_full_name(user_data):
-    # If no name values are defined on a user, the entire `name` object is not returned in the
-    # user data response; see https://github.com/CVEProject/cve-services/issues/436.
-    name = user_data.get("name", {})
-    if name:
-        return f"{name.get('first', '')} {name.get('last', '')}".strip() or None
+def get_full_name(user_data: dict) -> Optional[str]:
+    name_data = user_data["name"]
+    return f"{name_data.get('first', '')} {name_data.get('last', '')}".strip() or None
 
 
-def bool_to_text(value):
+def bool_to_text(value: Optional[bool]) -> str:
     if value is None:
         return "N/A"
     return "Yes" if value else "No"
 
 
-def natural_cve_sort(cve):
+def natural_cve_sort(cve: str) -> List[int]:
     if not cve:
         return []
     return [int(x) for x in cve.split("-")[1:]]
 
 
-def handle_cve_api_error(func):
+def handle_cve_api_error(func: Callable) -> Callable:
     """Decorator for catching CVE API exceptions and formatting the error message."""
 
     @wraps(func)
-    def wrapped(*args, **kwargs):
+    def wrapped(*args: Any, **kwargs: Any) -> Callable:
         try:
             return func(*args, **kwargs)
-        except CveApiError as exc:
-            error, _, details = str(exc).partition("; returned error: ")
-            click.secho("ERROR: ", bold=True, nl=False)
-            click.echo(error)
-            if details:
-                click.secho("DETAILS: ", bold=True, nl=False)
-                click.echo(details)
-            sys.exit(1)
+        except requests.exceptions.RequestException as exc:
+            error = str(exc)
+            details = None
+            if getattr(exc, "response", None) is not None:
+                try:
+                    details = exc.response.json()
+                except ValueError:
+                    details = exc.response.content
+
+        click.secho("ERROR: ", bold=True, nl=False)
+        click.echo(error)
+        if details:
+            click.secho("DETAILS: ", bold=True, nl=False)
+            click.echo(details)
+        sys.exit(1)
 
     return wrapped
 
 
 class Config:
-    def __init__(self, username, org, api_key, env, api_url, interactive):
+    def __init__(
+        self,
+        username: str,
+        org: str,
+        api_key: str,
+        env: str,
+        api_url: Optional[str],
+        interactive: bool,
+    ) -> None:
         self.username = username
         self.org = org
         self.api_key = api_key
@@ -134,7 +157,7 @@ class Config:
         self.interactive = interactive
         self.cve_api = self.init_cve_api()
 
-    def init_cve_api(self):
+    def init_cve_api(self) -> CveApi:
         return CveApi(
             username=self.username,
             org=self.org,
@@ -189,11 +212,189 @@ class Config:
     __version__, "-V", "--version", prog_name="cvelib", message="%(prog)s %(version)s"
 )
 @click.pass_context
-def cli(ctx, username, org, api_key, env, api_url, interactive):
+def cli(
+    ctx: click.Context,
+    username: str,
+    org: str,
+    api_key: str,
+    env: str,
+    api_url: Optional[str],
+    interactive: bool,
+) -> None:
     """A CLI interface for the CVE Services API."""
     ctx.obj = Config(username, org, api_key, env, api_url, interactive)
 
 
+@cli.command()
+@click.argument("cve_id", callback=validate_cve)
+@click.option(
+    "-j",
+    "--cve-json",
+    "cve_json_str",
+    required=True,
+    type=click.STRING,
+    help="JSON body of CVE record to publish.",
+)
+@click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
+@click.pass_context
+@handle_cve_api_error
+def publish(ctx: click.Context, cve_id: str, cve_json_str: str, print_raw: bool) -> None:
+    """Publish a CVE record for a reserved (or rejected) CVE ID.
+
+    If the CVE is already published, this action will update its record. A published CVE can only be
+    moved to the rejected state with an appropriate reject record (see `cve reject`). A published
+    CVE cannot be moved back to the reserved state.
+
+    Example:
+
+    \b
+    cve publish CVE-2022-1234 -j '{"affected": [], "descriptions": [], "references": {}, ...}'
+
+    For information on the required properties in a given CVE JSON record, see the
+    `cnaPublishedContainer` schema in:\n
+    https://github.com/CVEProject/cve-schema/blob/master/schema/v5.0/CVE_JSON_5.0_schema.json
+    """
+    try:
+        cve_json = json.loads(cve_json_str)
+    except json.JSONDecodeError as exc:
+        click.echo("CVE data was not valid JSON. Error was:\n")
+        click.secho(str(exc))
+        return
+    if ctx.obj.interactive:
+        click.echo("You are about to publish a CVE record for ", nl=False)
+        click.secho(cve_id, bold=True, nl=False)
+        click.echo(" using the following input:\n\n", nl=False)
+        click.secho(cve_json_str, bold=True, nl=False)
+        if not click.confirm("\n\nDo you want to continue?"):
+            click.echo("Exiting...")
+            sys.exit(0)
+        click.echo()
+
+    cve_api = ctx.obj.cve_api
+    try:
+        response_data = cve_api.publish(cve_id, cve_json)
+        created = True
+    except requests.exceptions.HTTPError as exc:
+        error = exc.response.json()["error"]
+        if exc.response.status_code != 403 or error != cve_api.Errors.RECORD_EXISTS:
+            raise exc
+        response_data = cve_api.update_published(cve_id, cve_json)
+        created = False
+    if print_raw:
+        print_json_data(response_data)
+    else:
+        click.echo("Published the following CVE:\n")
+        print_cve_record(response_data["created"] if created else response_data["updated"])
+
+
+@cli.command()
+@click.argument("cve_id", callback=validate_cve)
+@click.option(
+    "-j",
+    "--cve-json",
+    "cve_json_str",
+    type=click.STRING,
+    help="JSON body of CVE record to reject.",
+)
+@click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
+@click.pass_context
+@handle_cve_api_error
+def reject(ctx: click.Context, cve_id: str, cve_json_str: str, print_raw: bool) -> None:
+    """Reject a CVE record for a reserved or published CVE ID.
+
+    If the CVE is already rejected, this action will update its record if one is supplied.
+    A rejected CVE with a record can only be moved to the published state (see `cve publish`).
+    A rejected CVE without a record can be moved to the reserved state. A published CVE can only
+    be rejected with an accompanying record. Reserved CVEs can be rejected with or without a record.
+
+    Example:
+
+    \b
+    cve reject CVE-2022-1234 -j '{"rejectedReasons": [{"lang": "en", "value": "A reason."}]}'
+
+    For information on the required properties in a given CVE JSON record, see the
+    `cnaRejectedContainer` schema in:\n
+    https://github.com/CVEProject/cve-schema/blob/master/schema/v5.0/CVE_JSON_5.0_schema.json
+    """
+    cve_json = None
+    if cve_json_str:
+        try:
+            cve_json = json.loads(cve_json_str)
+        except json.JSONDecodeError as exc:
+            click.echo("CVE data was not valid JSON. Error was:\n")
+            click.secho(str(exc))
+            return
+
+    if ctx.obj.interactive:
+        click.echo("You are about to reject ", nl=False)
+        click.secho(cve_id, bold=True, nl=False)
+        if cve_json:
+            click.echo(" using the following input:\n\n", nl=False)
+            click.secho(cve_json_str)
+        else:
+            click.echo(" without providing a reject record.")
+        if not click.confirm("\nDo you want to continue?"):
+            click.echo("Exiting...")
+            sys.exit(0)
+        click.echo()
+
+    cve_api = ctx.obj.cve_api
+
+    # Reject a CVE ID without a record
+    if cve_json is None:
+        response_data = cve_api.move_to_rejected(cve_id)
+        if print_raw:
+            print_json_data(response_data)
+        else:
+            click.echo("Rejected the following CVE:\n")
+            print_cve_id(response_data["updated"])
+        return
+
+    # Reject a CVE ID with a record
+    try:
+        response_data = cve_api.reject(cve_id, cve_json)
+        created = True
+    except requests.exceptions.HTTPError as exc:
+        error = exc.response.json()["error"]
+        if exc.response.status_code != 400 or error != cve_api.Errors.RECORD_EXISTS:
+            raise exc
+        response_data = cve_api.update_rejected(cve_id, cve_json)
+        created = False
+
+    if print_raw:
+        print_json_data(response_data)
+    else:
+        click.echo("Rejected the following CVE:\n")
+        print_cve_record(response_data["created"] if created else response_data["updated"])
+
+
+@cli.command()
+@click.argument("cve_id", callback=validate_cve)
+@click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
+@click.pass_context
+@handle_cve_api_error
+def undo_reject(ctx: click.Context, cve_id: str, print_raw: bool) -> None:
+    """Move a rejected CVE ID without a record back to the reserved state."""
+    if ctx.obj.interactive:
+        click.echo("You are about to move ", nl=False)
+        click.secho(cve_id, bold=True, nl=False)
+        click.echo(" back to the reserved state.", nl=False)
+        if not click.confirm(
+            "\nThis is only allowed for CVE IDs without a record. Do you want to continue?"
+        ):
+            click.echo("Exiting...")
+            sys.exit(0)
+        click.echo()
+
+    cve_api = ctx.obj.cve_api
+    response_data = cve_api.move_to_reserved(cve_id)
+    if print_raw:
+        print_json_data(response_data)
+    else:
+        click.echo("Moved the following CVE to reserved:\n")
+        print_cve_id(response_data["updated"])
+
+
 @cli.command()
 @click.option(
     "-r",
@@ -215,7 +416,7 @@ def cli(ctx, username, org, api_key, env, api_url, interactive):
 @click.argument("count", default=1, type=click.IntRange(min=1))
 @click.pass_context
 @handle_cve_api_error
-def reserve(ctx, random, year, count, print_raw):
+def reserve(ctx: click.Context, random: bool, year: str, count: int, print_raw: bool) -> None:
     """Reserve one or more CVE IDs. COUNT is the number of CVEs to reserve; defaults to 1.
 
     CVE IDs can be reserved one by one (the lowest IDs are reserved first) or in batches of
@@ -223,7 +424,9 @@ def reserve(ctx, random, year, count, print_raw):
     generated sequentially (default) or non-sequentially (random IDs are selected from your CVE ID
     range).
 
-    For more information, see: "Developer Guide to CVE Services API" (https://git.io/JLcmZ)
+    \b
+    For more information, see the "Developer Guide to CVE Services API":
+    https://github.com/CVEProject/cve-services/wiki/Developer-Guide-to-CVE-Services-API#different-reservation-types
     """
     if random and count > 10:
         raise click.BadParameter("requesting non-sequential CVE IDs is limited to 10 per request.")
@@ -248,45 +451,75 @@ def reserve(ctx, random, year, count, print_raw):
         click.echo()
 
     cve_api = ctx.obj.cve_api
-    cve_data, remaining_quota = cve_api.reserve(count, random, year)
+    cve_data = cve_api.reserve(count, random, year)
     if print_raw:
         print_json_data(cve_data)
         return
 
     click.echo("Reserved the following CVE ID(s):\n")
-    for cve in cve_data["cve_ids"]:
-        print_cve(cve)
+    for cve_id_data in cve_data["cve_ids"]:
+        print_cve_id(cve_id_data)
 
-    click.echo(f"\nRemaining quota: {remaining_quota}")
+    click.echo(f"\nRemaining quota: {cve_data['meta']['remaining_quota']}")
 
 
 @cli.command(name="show")
+@click.option(
+    "-r",
+    "--show-record",
+    default=False,
+    is_flag=True,
+    help="Show full CVE record in JSON v5 format.",
+)
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.argument("cve_id", callback=validate_cve)
 @click.pass_context
 @handle_cve_api_error
-def show_cve(ctx, print_raw, cve_id):
-    """Display a specific CVE ID owned by your CNA."""
+def show_cve(ctx: click.Context, show_record: bool, print_raw: bool, cve_id: str) -> None:
+    """Display a specific CVE ID (and optionally its record) owned by your CNA."""
     cve_api = ctx.obj.cve_api
-    cve = cve_api.show_cve(cve_id=cve_id)
+
+    cve_id_data = cve_api.show_cve_id(cve_id=cve_id)
+    cve_record_data = {}
+    if show_record:
+        try:
+            cve_record_data = cve_api.show_cve_record(cve_id=cve_id)
+        except requests.exceptions.HTTPError as exc:
+            error_msg = exc.response.json()["error"]
+            if exc.response.status_code != 404 or error_msg != cve_api.Errors.RECORD_DOES_NOT_EXIST:
+                raise exc
+
     if print_raw:
-        print_json_data(cve)
+        # Serialize the CVE ID data and CVE record into an array for easier parsing if we're
+        # showing the record as well. If not, just display the CVE ID data.
+        if show_record:
+            print_json_data([cve_id_data, cve_record_data])
+        else:
+            print_json_data(cve_id_data)
     else:
-        print_cve(cve)
+        print_cve_id(cve_id_data)
+        # If we're showing the CVE record, display it as either the raw JSON if it exists or show
+        # an informational message if it does not.
+        if show_record:
+            click.secho("-----", bold=True)
+            if cve_record_data:
+                print_json_data(cve_record_data)
+            else:
+                click.echo("CVE record does not exist.")
 
 
 @cli.command(name="list")
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.option(
     "--sort-by",
-    type=click.Choice(["cve_id", "state", "user", "reserved"], case_sensitive=False),
+    type=click.Choice(["cve_id", "state", "user", "reserved_ts"], case_sensitive=False),
     default="cve_id",
     help="Sort output.",
 )
 @click.option("--year", callback=validate_year, help="Filter by year.")
 @click.option(
     "--state",
-    type=click.Choice(["reserved", "published", "rejected"], case_sensitive=False),
+    type=click.Choice(CveApi.States.values(), case_sensitive=False),
     help="Filter by reservation state.",
 )
 @click.option(
@@ -297,7 +530,7 @@ def show_cve(ctx, print_raw, cve_id):
 )
 @click.pass_context
 @handle_cve_api_error
-def list_cves(ctx, print_raw, sort_by, **query):
+def list_cves(ctx: click.Context, print_raw: bool, sort_by: str, **query: dict) -> None:
     """Filter and list reserved CVE IDs owned by your CNA."""
     cve_api = ctx.obj.cve_api
     cves = list(cve_api.list_cves(**query))
@@ -315,12 +548,12 @@ def list_cves(ctx, print_raw, sort_by, **query):
             cves.sort(key=lambda x: x["requested_by"]["user"])
         elif key == "cve_id":
             cves.sort(key=lambda x: natural_cve_sort(x["cve_id"]))
-        elif key == "reserved_asc":
+        elif key == "reserved_ts":
             cves.sort(key=lambda x: x["reserved"])
         elif key == "state":
             cves.sort(key=lambda x: x["state"])
 
-    lines = [("CVE ID", "STATE", "OWNING CNA", "REQUESTED BY", "RESERVED")]
+    lines = [("CVE ID", "STATE", "OWNING CNA", "RESERVED BY", "RESERVED ON")]
     for cve in cves:
         lines.append(
             (
@@ -338,7 +571,7 @@ def list_cves(ctx, print_raw, sort_by, **query):
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.pass_context
 @handle_cve_api_error
-def quota(ctx, print_raw):
+def quota(ctx: click.Context, print_raw: bool) -> None:
     """Display the available CVE ID quota for your CNA.
 
     \b
@@ -370,7 +603,7 @@ def quota(ctx, print_raw):
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.pass_context
 @handle_cve_api_error
-def show_user(ctx, username, print_raw):
+def show_user(ctx: click.Context, username: Optional[str], print_raw: bool) -> None:
     """Show information about a user."""
     if ctx.invoked_subcommand is not None:
         return
@@ -396,7 +629,7 @@ def show_user(ctx, username, print_raw):
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.pass_context
 @handle_cve_api_error
-def reset_key(ctx, username, print_raw):
+def reset_key(ctx: click.Context, username: Optional[str], print_raw: bool) -> None:
     """Reset a user's personal access token (API key).
 
     This API key is used to authenticate each request to the CVE API.
@@ -435,7 +668,7 @@ def reset_key(ctx, username, print_raw):
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.pass_context
 @handle_cve_api_error
-def update_user(ctx, username, **opts_data):
+def update_user(ctx: click.Context, username: Optional[str], **opts_data: dict) -> None:
     """Update a user.
 
     To reset a user's API key, use `cve user reset-key`.
@@ -454,7 +687,7 @@ def update_user(ctx, username, **opts_data):
                 opt = "active_roles." + opt.replace("_role", "")
             elif opt == "active":
                 # Convert boolean to string since this data is passed as query params
-                value = str(value).lower()
+                value = str(value).lower()  # type: ignore
             user_updates[opt] = value
 
     if not user_updates:
@@ -489,7 +722,9 @@ def update_user(ctx, username, **opts_data):
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.pass_context
 @handle_cve_api_error
-def create_user(ctx, username, name_first, name_last, roles, print_raw):
+def create_user(
+    ctx: click.Context, username: str, name_first: str, name_last: str, roles: list, print_raw: bool
+) -> None:
     """Create a user in your organization.
 
     This action is restricted to users with the ADMIN role.
@@ -497,7 +732,7 @@ def create_user(ctx, username, name_first, name_last, roles, print_raw):
     Note: Once a user is created, they cannot be removed, only marked as inactive. Only create
     users when you really need them.
     """
-    user_data = defaultdict(dict)
+    user_data: DefaultDict = defaultdict(dict)
     user_data["username"] = username
 
     if name_first:
@@ -539,7 +774,7 @@ def create_user(ctx, username, name_first, name_last, roles, print_raw):
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.pass_context
 @handle_cve_api_error
-def show_org(ctx, print_raw):
+def show_org(ctx: click.Context, print_raw: bool) -> None:
     """Show information about your organization."""
     if ctx.invoked_subcommand is not None:
         return
@@ -560,7 +795,7 @@ def show_org(ctx, print_raw):
 @click.option("--raw", "print_raw", default=False, is_flag=True, help="Print response JSON.")
 @click.pass_context
 @handle_cve_api_error
-def users(ctx, print_raw):
+def users(ctx: click.Context, print_raw: bool) -> None:
     """List all users in your organization."""
     cve_api = ctx.obj.cve_api
     org_users = list(cve_api.list_users())
@@ -588,14 +823,16 @@ def users(ctx, print_raw):
 
 @cli.command()
 @click.pass_context
-def ping(ctx):
+@handle_cve_api_error
+def ping(ctx: click.Context) -> None:
     """Ping the CVE Services API to see if it is up."""
     cve_api = ctx.obj.cve_api
-    ok, error_msg = cve_api.ping()
+    error = cve_api.ping()
 
     click.echo(f"CVE API Status — {cve_api.url}\n└─ ", nl=False)
-    if ok:
-        click.secho("OK", fg="green")
-    else:
-        click.secho("ERROR:", bold=True, nl=False)
-        click.echo(f" {error_msg}")
+    if error:
+        # Raise the exception again so the decorator pretty-prints the output
+        # We don't directly raise in ping() so we can print the status line first
+        raise error
+    # else ping() returned None
+    click.secho("OK", fg="green")
